%% Matlab script for testing allocateProduction function, without having to run simulation.
% Conditions are set in test11.m
clear all;
test11; % load example data
fprintf('\n\n********** ALLOCATING PRODUCTION... ****************\n');

demand = sum(demands);
[n, m] = size(Q_max);

if demand < 0
    fprintf('Production DOWN\n');
    if sum(sum(Q_min)) > demand
        fprintf('NOT ENOUGH BIDS TO COVER DEMAND\n');
    end
else
    fprintf('Production UP\n');
    if sum(sum(Q_max)) < demand
        fprintf('NOT ENOUGH BIDS TO COVER DEMAND\n');
    end
end

fprintf('***** Finding solution... ***** \n');

% Decision variables
onoff = binvar(n,m,'full');
P = sdpvar(n,m,'full');
prod_slack = sdpvar(n,1,'full');
delta = sdpvar(n,1,'full');
delta(1) = 0; % Phase angle reference

% Help variables
flow = D*A*delta;

prod = sum(P,2);

% Set contraints
matchTotalDemand = sum(prod) == demand;
noOverproduction = prod_slack >= 0; 
bidSizeMin = onoff .* Q_min <= P;
bidSizeMax = P <= onoff .* Q_max;
findAngles = prod - demands == B_prime*delta;
findAnglesWithSlack = prod - demands + prod_slack == B_prime*delta;
flowCapMin = flow >= minFlowVector;
flowCapMax = flow <= maxFlowVector;

Constraints = [matchTotalDemand bidSizeMin bidSizeMax findAngles flowCapMin flowCapMax];

% Set objectives
powerCost = sum(sum(C .* abs(P)));
clogCost = flowCost*sum(abs(flow));
prodSlackCost = sum(prod_slack)*10000;
activationCost = activationCost*sum(sum(onoff));

Objective = powerCost + clogCost + activationCost;

% Solve the problem
%options = sdpsettings('solver','INTLINPROG');
options = sdpsettings('solver','cplex');
res = optimize(Constraints,Objective, options)
no_solution = res.problem;

% If no feasible solution, get best effort solution
if no_solution
    fprintf('\n ***** Failed to find feasible solution. Demand cannot be met. ***** \n');
    fprintf('***** Finding best-effort solution... ***** \n');
    Constraints = [noOverproduction bidSizeMin bidSizeMax findAnglesWithSlack flowCapMin flowCapMax];
    Objective = powerCost + clogCost + prodSlackCost;
    res = optimize(Constraints,Objective)
    no_solution = res.problem;
    if no_solution
        fprintf('***** ERROR: No solution. Applying zero FRR-A production  ***** \n');
    else
        fprintf('***** Best-effort solution found ***** \n');
    end
else
    fprintf('***** Solution found ***** \n');
    prod_slack = zeros(n,1);
end

% Store result in variables, set numbers smaller than 'eps' to zero
eps = 1e-8;
bids = value(P); bids(abs(bids)<eps)=0;
production = value(prod); production(abs(production)<eps)=0;
shortage = value(prod_slack); shortage(abs(shortage)<eps)=0;
cost.power = value(powerCost);
cost.clog = value(clogCost);
cost.activation = value(activationCost);
FRR_A_cost = cost.power;
export = production - demands;
flow_d = value(flow);

% If no solution, apply zero FRR-A production
if no_solution
    bids = zeros(n,m);
    production = zeros(n,1);
    shortage = demands;
    cost.power = 0;
    cost.clog = 0;
    cost.activation = 0;
    FRR_A_cost = cost.power;
    export = production - demands;
end

% Print to Matlab
prod_array = production';
flow_array = flow_d';
shortage_array = shortage';
prod_array
flow_array
shortage_array
cost
cost.total = cost.power + cost.clog + cost.activation;
fprintf('********** DONE ****************\n');